(* :Title: Tetris.m *)

(* :Author: Andrew Hunt, andy@wolfram.com *)

(* :Package Version: 1.00 *)

(* :Mathematica Version: 5.0.0 *)
                     
(* :Copyright: (c) 2004, Andrew Hunt, All rights reserved. *)

(* :Discussion: Tetris via Mathematica *)



BeginPackage["Tetris`", {"GUIKit`"}]

Tetris::usage = "Start a new Tetris game.";

TetrisGUI::usage = "TetrisGUI[] starts a GUI version of Tetris.";
$TetrisGUIObject::usage = "$TetrisGUIObject is the GUIObject of the current Tetris game";

(* Set default parameters *)
$boardWidth = 10;
$boardHeight = 20;
$pixelSize = 15;
$refreshTiming::usage = "$refreshTiming is the refresh timing rate.";

colorSet::usage = "The color settings used for pieces.";


(*PieceMovementLoop::usage = "";*)
$pieceOrientation::usage = "Orientation of Tetris piece";
$pieceOriginX = $newPieceOriginX = 1;
$pieceOriginY = $newPieceOriginY = 1;
$pieceStyle::usage = "Style of Tetris piece"; (* 1 === Square *)

$nextPieceStyle::usage = "Style of next Tetris piece";

getPieceCoordinates::usage = "getPieceCoordinates[x, y, shape, rotation] returns the list coordinates for the given shape at origin {x,y}.";
PiecePlacementQ::usage = "Test if Tetris piece can be placed on board";
getNextPiece::usage = "Numerical value of next piece";
getNextRotation::usage = "Numerical value of next rotation piece";

ElapsedTime::usage = "How much taime has elapsed?";
$refreshTiming::usage = "Refresh timing";
$currentLevel::usage = "Current player level";
$ActiveGameQ::usage = "Is game active?";
PieceCount::usage = "How many pieces on board?";
RowCount::usage = "How many rows?";
AddPieceToQueue::usage = "Add Tetris piece to queue";
CollapseRows::usage = "Collapse rows of matrix";
getPieceDimensions::usage = "Dimensions of Tetris piece";
defaultSettings::usage = "Default / opening settings for Tetris game"; 
setNextPiecePanel::usage = "Set the visual for next Tetris piece";
(**)
gameOverQ::usage = "";

(**)
clearBoard::usage = "Clear Tetris board";
(**)
newGame::usage = "";
(**)
StartTime::usage = "Time when Tetris game started.";

$statusMode::usage = "";

$TetrisBoard::usage = "Current state of Tetris game.";
TetrisPlot::usage = "TetrisPlot[board] plots board using ListDensityPlot.";



Begin["`Private`"]


$TetrisDirectory = DirectoryName[ System`Private`$InputFileName];

(* Set debug mode *)
$debug = False;
$statusMode = True;

(* Level of play *)
$currentLevel = 0;
(* Timing between steps *)
$refreshTiming = IntegerPart[1000 * ((10.0 - $currentLevel) / 20.0)];
(* Is game active? *)
$ActiveGameQ = False;
(* Keep track of number of pieces played *)
PieceCount = 0;
RowCount = 0;
(* Default piece style *)
$pieceStyle = 3;

		
(* Set extra top space for pieces to start within *)
topSpace = 4;

(* Get new piece *)
getNextPiece[]:= Random[Integer, {1, 7}];

(* *)
clearBoard[]:= ($TetrisBoard = Table[0, {$boardWidth}, {$boardHeight}] );

(* start new game *)
newGame[]:= newGame[10, 20, 15];

newGame[bw_, bh_, px_]:= 
	(
		$boardWidth = ToExpression[bw];
		$boardHeight = ToExpression[bh];
		$pixelSize = ToExpression[px];

		clearBoard[];
		clearScreen[];
		PieceCount = 0;
		RowCount = 0;
		$nextPieceStyle = 0;

		$ActiveGameQ = True;
		defaultSettings[1];
		AddPieceToQueue[];

		StartTime = AbsoluteTime[];
		$TetrisBoard = Table[0, {$boardWidth}, {$boardHeight}];

		(* Future Delvelopment *)
		$currentAction = Null;
		$currentUser = Null;
		$previousAction = Null;
		$previousUser = Null;
	);


(* New Game *)
Tetris[]:=
	(
		newGame[];
		TetrisGUI[];
	);

TetrisGUI[] := (
	If[! GUIObjectQ[$TetrisGUIObject],
	$TetrisGUIObject = GUILoad["Tetris/Game"]]; 
	$TetrisGUIObject = GUIRun[$TetrisGUIObject, ReleaseMethod -> Manual]
	);


(* Main Loop *)
PieceMovementLoop[]:=
	Module[{lst},
		If[(($ActiveGameQ === True) && ($pieceOriginY <= $boardHeight-5)),
			(* Remove current piece *)
			lst = getPieceCoordinates[$pieceOriginX, $pieceOriginY, $pieceStyle, $pieceOrientation];
(* 			If[$debug === True, Print[lst] ]; *)
			InvokeMethod[{"gamePanel", "fillImagePixelArray"}, lst, 0, False, InvokeThread -> "Dispatch"];
			(Part[$TetrisBoard, #[[1]], #[[2]]] = 0) & /@ lst;
			(* Test new position *)
			If[PiecePlacementQ[$TetrisBoard, $pieceOriginX, $pieceOriginY, 
				$pieceStyle, $pieceOrientation] === True, 				
				(* Paint new piece *)
				$pieceOriginY++;
				lst = getPieceCoordinates[$pieceOriginX, $pieceOriginY, $pieceStyle, $pieceOrientation];
				If[$debug === True, Print["{", $pieceOriginX, ", ", $pieceOriginY, "} ", $pieceStyle, " ", $pieceOrientation] ];
				InvokeMethod[{"gamePanel", "fillImagePixelArray"}, lst, $pieceStyle, True, InvokeThread -> "Dispatch"]
			];
		,
			$ActiveGameQ = False;
		];
	];



(*  Game over *)
GameOver[]:= 
	(
		InvokeMethod[{"moveTimer", "stop"}];
		Print["Elapsed Time: " <> ToString[ ElapsedTime[StartTime, AbsoluteTime[]] ] ];
	);

gameOverQ[]:=
	Plus @@ (If[(Length@Select[#, (# > 0) &] > 0), 1, 0] & /@
		Drop[Transpose[$TetrisBoard], 1]) >= $boardHeight - 3;


(* Test if piece can be placed on board *)
PiecePlacementQ[board_, x_, y_, piece_, rot_]:=
	Module[{lst, t, xDim, yDim},
		lst = getPieceCoordinates[x, y, piece, rot];
		If[lst === $Failed, Return[False]];

		(* Test if point if off board *)
		t = Transpose[lst];
		{xDim, yDim} = Dimensions[board];

		If[Length@Select[First[t], (xDim-1 < #) &] > 0, Return[False]];
		If[Length@Select[Flatten@Rest[t], (yDim-1 < #) &] > 0, Return[False]];
		If[Length@Select[First[t], (0 > #) &] > 0, Return[False]];
		If[Length@Select[Flatten@Rest[t], (0 > #) &] > 0, Return[False]];

		If[$debug === True, Print[board, ", ", lst]];
		If[Plus @@ (Part[board, #[[1]]+1, #[[2]]+1] & /@ lst) > 0, False, True] ];



(* Compute elapsed time in seconds *)
ElapsedTime[start_, end_]:= (end - start);



(* Add piece to queue *)
AddPieceToQueue[]:=
	Module[{x, y},
		PieceCount++;
		(* Get next piece *)
		If[$nextPieceStyle === 0, 
			$pieceStyle = getNextPiece[];
			$nextPieceStyle = getNextPiece[];
		,
			$pieceStyle = $nextPieceStyle;
			$nextPieceStyle = getNextPiece[];
		];
		$pieceOrientation = 1;
		{x, y} = getPieceDimensions[$pieceStyle, $pieceOrientation];
		$pieceOriginX = Ceiling[($boardWidth - x)/2]+1;
		$pieceOriginY = 1;
		setNextPiecePanel[];
	];


(* Add piece to queue *)
setNextPiecePanel[]:=
		($TetrisGUIObject @ SetPropertyValue[{"nextPiecePanel", "imagePixels"},
			Flatten[ Table[0, {5}, {3}]], InvokeThread -> "Dispatch"];
		$TetrisGUIObject @ InvokeMethod[{"nextPiecePanel", "fillImagePixelArray"},
			getPieceCoordinates[2, 1, $nextPieceStyle, 1], $nextPieceStyle, True, 
			InvokeThread -> "Dispatch"];);



(* Collapse Rows *)
CollapseRows[]:=
	Module[{lst, rowNums, i, tempBoard = Transpose[$TetrisBoard]},
		(* Gather Rows to delete *)
		lst = Select[tempBoard, !MemberQ[#, 0]&];
		If[Length[lst] > 0,
			(* Stop Timer *)
			$TetrisGUIObject @ InvokeMethod[{"moveTimer", "stop"}];
			rowNums = Flatten[Position[tempBoard, #] & /@ lst];
			(RowCount++;
			tempBoard = Drop[tempBoard, {#}]) & /@ 
				(rowNums + Table[i, {i, 0, -1*Length[lst] + 1, -1}]);
			tempBoard = Join[Table[0, {Length[rowNums]}, {10}], tempBoard];
			$TetrisBoard = Transpose[tempBoard];
			$TetrisGUIObject @ 
				SetPropertyValue[{"gamePanel", "imagePixels"}, Flatten[Transpose@$TetrisBoard]];
			(* Start Timer *)
			$TetrisGUIObject @ InvokeMethod[{"moveTimer", "start"}];
		]
	];



(* Compute height of a column *)
getColumnHeight[x_]:= 0;


(* get piece width and height *)
(* TODO: Needs to be more dynamic *)
getPieceDimensions[s_, r_]:=
	Switch[s,
		1, Part[{{2,2}}, r],
		2, Part[{{1,4}, {4,1}}, r],
		3, Part[{{4,2}, {2,4}}, r],
		4, Part[{{4,2}, {2,4}}, r],
		5, Part[{{3,2}, {2,3}, {3,2}, {2,3}}, r],
		6, Part[{{3,2}, {2,3}, {3,2}, {2,3}}, r],
		7, Part[{{3,2}, {2,3}, {3,2}, {2,3}}, r]
	];


getNextRotation[int_Integer]:= If[ int > 3, 1, int + 1];
getNextRotation[int_]:= Print["Rotation problem: ", int, " Head:", Head@int];


(* *)
defaultSettings[]:= defaultSettings[1];

defaultSettings[int_]:=
	(
		$pieceStyle = int; 
		$pieceOriginX = Ceiling[$boardWidth/2]; 
		$pieceOriginY = 1;
		$pieceOrientation = 1;
	);


(* Color settings for pieces *)
colorSet[]:= 
{
	(* Black *) {0, 0, 0},
	(* Blue *)  {128, 126, 253},
	(* Green *)  {130, 255, 124},
	(* Orange *)  {255, 158, 63},
	(* Purple *)  {253, 124, 253},
	(* Red *)  {253, 63, 70},
	(* Teal *)  {124, 255, 251},
	(* Yellow *)  {251, 253, 59}
};


(* Custom ColorFunction from Jeff 'Java Jeff' Adams *)
TetrisPlot[board_] :=
	Module[{cls = Apply[RGBColor, colorSet[]/255., {1}]},
		ListDensityPlot[ Reverse[Transpose[board]],
			Frame -> False,
			Mesh -> True,
			MeshStyle -> {RGBColor[0.4, 0.4, 0.4]},
			ColorFunctionScaling -> False,
			ColorFunction -> (Part[cls, # + 1] &), AspectRatio -> Automatic]
	];



(* Handle piece translation and rotation *)
getPieceCoordinates[x_Integer, y_Integer, shapeNum_Integer, rotation_Integer]:= 
	Module[{pos},
		If[($boardWidth < x < 1) || ($boardHeight < y < 1), Return[$Failed] ];
		pos = Switch[shapeNum,
			1, (* 'O' *)
				{ {x-1, y}, {x, y}, {x-1, y+1}, {x, y+1} },
			2, (* 'I' *)
				Part[{
					{ {x-2, y}, {x-1, y}, {x, y}, {x+1, y} },
					{ {x, y-1}, {x, y}, {x, y+1}, {x, y+2} },
					{ {x-2, y}, {x-1, y}, {x, y}, {x+1, y} },
					{ {x, y-1}, {x, y}, {x, y+1}, {x, y+2} }
				}, rotation],
			3, (* 'S' *)
				Part[{
					{ {x-1, y}, {x, y}, {x, y-1}, {x+1, y-1} },
					{ {x, y-1}, {x, y}, {x+1, y}, {x+1, y+1} },
					{ {x-1, y}, {x, y}, {x, y-1}, {x+1, y-1} },
					{ {x, y-1}, {x, y}, {x+1, y}, {x+1, y+1} }
				}, rotation],
			4, (* 'Z' *)
				Part[{
					{ {x, y}, {x+1, y}, {x-1, y-1}, {x, y-1} },
					{ {x, y+1}, {x, y}, {x+1, y}, {x+1, y-1} },
					{ {x, y}, {x+1, y}, {x-1, y-1}, {x, y-1} },
					{ {x, y+1}, {x, y}, {x+1, y}, {x+1, y-1} }
				}, rotation],
			5, (* 'L' *)
				Part[{
					{ {x-1, y}, {x-1, y+1}, {x, y}, {x+1, y} },
					{ {x, y+1}, {x, y}, {x, y-1}, {x+1, y+1} },
					{ {x-1, y}, {x, y}, {x+1, y}, {x+1, y-1} },
					{ {x-1, y-1}, {x, y-1}, {x, y}, {x, y+1} }
				}, rotation],
			6, (* 'J' *)
				Part[{
					{ {x-1, y}, {x, y}, {x+1, y}, {x+1, y+1} },
					{ {x, y-1}, {x, y}, {x, y+1}, {x+1, y-1} },
					{ {x-1, y-1}, {x-1, y}, {x, y}, {x+1, y} },
					{ {x, y-1}, {x, y}, {x-1, y+1}, {x, y+1} }
				}, rotation],
			7, (* 'T' *)
				Part[{
					{ {x-1, y}, {x, y}, {x+1, y}, {x, y+1} },
					{ {x, y-1}, {x, y}, {x+1, y}, {x, y+1} },
					{ {x-1, y}, {x, y-1}, {x, y}, {x+1, y} },
					{ {x, y-1}, {x-1, y}, {x, y}, {x, y+1} }
				}, rotation]
]];




End[];

EndPackage[]

